@@ -49,11 +49,11 @@
 #define APFS_OBJECT_TYPE_GBITMAP		0x00000019
 #define APFS_OBJECT_TYPE_GBITMAP_TREE		0x0000001a
 #define APFS_OBJECT_TYPE_GBITMAP_BLOCK		0x0000001b
-#define OBJECT_TYPE_ER_RECOVERY_BLOCK		0x0000001c
-#define OBJECT_TYPE_SNAP_META_EXT		0x0000001d
-#define OBJECT_TYPE_INTEGRITY_META		0x0000001e
-#define OBJECT_TYPE_FEXT_TREE			0x0000001f
-#define OBJECT_TYPE_RESERVED_20			0x00000020
+#define APFS_OBJECT_TYPE_ER_RECOVERY_BLOCK	0x0000001c
+#define APFS_OBJECT_TYPE_SNAP_META_EXT		0x0000001d
+#define APFS_OBJECT_TYPE_INTEGRITY_META		0x0000001e
+#define APFS_OBJECT_TYPE_FEXT_TREE		0x0000001f
+#define APFS_OBJECT_TYPE_RESERVED_20		0x00000020
 #define APFS_OBJECT_TYPE_INVALID		0x00000000
 #define APFS_OBJECT_TYPE_TEST			0x000000ff
 
