@@ -81,24 +81,32 @@ struct apfs_node *apfs_read_node(struct super_block *sb, u64 oid, u32 storage,
 	case APFS_OBJ_VIRTUAL:
 		/* All virtual nodes are inside a volume, at least for now */
 		err = apfs_omap_lookup_block(sb, sbi->s_omap, oid, &bno, write);
-		if (err)
+		if (err) {
+			apfs_err(sb, "omap lookup failed for oid 0x%llx", oid);
 			return ERR_PTR(err);
+		}
 		/* CoW has already been done, don't worry about snapshots */
 		bh = apfs_read_object_block(sb, bno, write, false /* preserve */);
-		if (IS_ERR(bh))
+		if (IS_ERR(bh)) {
+			apfs_err(sb, "object read failed for bno 0x%llx", bno);
 			return (void *)bh;
+		}
 		break;
 	case APFS_OBJ_PHYSICAL:
 		bh = apfs_read_object_block(sb, oid, write, false /* preserve */);
-		if (IS_ERR(bh))
+		if (IS_ERR(bh)) {
+			apfs_err(sb, "object read failed for bno 0x%llx", oid);
 			return (void *)bh;
+		}
 		oid = bh->b_blocknr;
 		break;
 	case APFS_OBJ_EPHEMERAL:
 		/* Ephemeral objects are checkpoint data, so ignore 'write' */
 		bh = apfs_read_ephemeral_object(sb, oid);
-		if (IS_ERR(bh))
+		if (IS_ERR(bh)) {
+			apfs_err(sb, "ephemeral read failed for oid 0x%llx", oid);
 			return (void *)bh;
+		}
 		break;
 	}
 	raw = (struct apfs_btree_node_phys *) bh->b_data;
@@ -130,12 +138,12 @@ struct apfs_node *apfs_read_node(struct super_block *sb, u64 oid, u32 storage,
 
 	if (nxi->nx_flags & APFS_CHECK_NODES && !apfs_obj_verify_csum(sb, bh)) {
 		/* TODO: don't check this twice for virtual/physical objects */
-		apfs_alert(sb, "bad checksum for node in block 0x%llx", bh->b_blocknr);
+		apfs_err(sb, "bad checksum for node in block 0x%llx", bh->b_blocknr);
 		apfs_node_free(node);
 		return ERR_PTR(-EFSBADCRC);
 	}
 	if (!apfs_node_is_valid(sb, node)) {
-		apfs_alert(sb, "bad node in block 0x%llx", bh->b_blocknr);
+		apfs_err(sb, "bad node in block 0x%llx", bh->b_blocknr);
 		apfs_node_free(node);
 		return ERR_PTR(-EFSCORRUPTED);
 	}
@@ -167,6 +175,16 @@ static int apfs_node_min_table_size(struct super_block *sb, u32 type, u16 flags)
 		val_size = sizeof(__le64); /* We assume no ghosts here */
 		toc_size = sizeof(struct apfs_kvoff);
 		break;
+	case APFS_OBJECT_TYPE_OMAP_SNAPSHOT:
+		key_size = sizeof(__le64);
+		val_size = leaf ? sizeof(struct apfs_omap_snapshot) : sizeof(__le64);
+		toc_size = sizeof(struct apfs_kvoff);
+		break;
+	case APFS_OBJECT_TYPE_FEXT_TREE:
+		key_size = sizeof(struct apfs_fext_tree_key);
+		val_size = leaf ? sizeof(struct apfs_fext_tree_val) : sizeof(__le64);
+		toc_size = sizeof(struct apfs_kvoff);
+		break;
 	default:
 		/* Make room for one record at least */
 		toc_size = sizeof(struct apfs_kvloc);
@@ -234,10 +252,13 @@ int apfs_make_empty_btree_root(struct super_block *sb, u32 subtype, u64 *oid)
 	ASSERT(subtype == APFS_OBJECT_TYPE_BLOCKREFTREE || subtype == APFS_OBJECT_TYPE_OMAP_SNAPSHOT);
 
 	err = apfs_spaceman_allocate_block(sb, &bno, true /* backwards */);
-	if (err)
+	if (err) {
+		apfs_err(sb, "block allocation failed");
 		return err;
+	}
 	apfs_assert_in_transaction(sb, &vsb_raw->apfs_o);
 	le64_add_cpu(&vsb_raw->apfs_fs_alloc_count, 1);
+	le64_add_cpu(&vsb_raw->apfs_total_blocks_alloced, 1);
 
 	bh = apfs_getblk(sb, bno);
 	if (!bh)
@@ -313,22 +334,32 @@ static struct apfs_node *apfs_create_node(struct super_block *sb, u32 storage)
 	switch (storage) {
 	case APFS_OBJ_VIRTUAL:
 		err = apfs_spaceman_allocate_block(sb, &bno, true /* backwards */);
-		if (err)
+		if (err) {
+			apfs_err(sb, "block allocation failed");
 			return ERR_PTR(err);
+		}
+		apfs_assert_in_transaction(sb, &vsb_raw->apfs_o);
 		le64_add_cpu(&vsb_raw->apfs_fs_alloc_count, 1);
+		le64_add_cpu(&vsb_raw->apfs_total_blocks_alloced, 1);
 
 		oid = le64_to_cpu(msb_raw->nx_next_oid);
 		le64_add_cpu(&msb_raw->nx_next_oid, 1);
 		err = apfs_create_omap_rec(sb, oid, bno);
-		if (err)
+		if (err) {
+			apfs_err(sb, "omap rec creation failed (0x%llx-0x%llx)", oid, bno);
 			return ERR_PTR(err);
+		}
 		break;
 	case APFS_OBJ_PHYSICAL:
 		err = apfs_spaceman_allocate_block(sb, &bno, true /* backwards */);
-		if (err)
+		if (err) {
+			apfs_err(sb, "block allocation failed");
 			return ERR_PTR(err);
+		}
 		/* We don't write to the container's omap */
+		apfs_assert_in_transaction(sb, &vsb_raw->apfs_o);
 		le64_add_cpu(&vsb_raw->apfs_fs_alloc_count, 1);
+		le64_add_cpu(&vsb_raw->apfs_total_blocks_alloced, 1);
 		oid = bno;
 		break;
 	case APFS_OBJ_EPHEMERAL:
@@ -337,8 +368,10 @@ static struct apfs_node *apfs_create_node(struct super_block *sb, u32 storage)
 		le64_add_cpu(&msb_raw->nx_next_oid, 1);
 
 		err = apfs_create_cpoint_map(sb, oid, bno);
-		if (err)
+		if (err) {
+			apfs_err(sb, "checkpoint map creation failed (0x%llx-0x%llx)", oid, bno);
 			return ERR_PTR(err);
+		}
 		break;
 	default:
 		ASSERT(false);
@@ -414,42 +447,56 @@ int apfs_delete_node(struct apfs_query *query)
 	 * deleting the node, because that involves moving blocks around.
 	 */
 	err = apfs_btree_remove(query->parent);
-	if (err)
+	if (err) {
+		apfs_err(sb, "parent index removal failed for oid x%llx", oid);
 		return err;
+	}
 
 	switch (query->flags & APFS_QUERY_TREE_MASK) {
 	case APFS_QUERY_CAT:
 		err = apfs_free_queue_insert(sb, bno, 1);
-		if (err)
+		if (err) {
+			apfs_err(sb, "free queue insertion failed for 0x%llx", bno);
 			return err;
+		}
 		err = apfs_delete_omap_rec(sb, oid);
-		if (err)
+		if (err) {
+			apfs_err(sb, "omap rec deletion failed (0x%llx)", oid);
 			return err;
+		}
 		vsb_raw = APFS_SB(sb)->s_vsb_raw;
 		apfs_assert_in_transaction(sb, &vsb_raw->apfs_o);
 		le64_add_cpu(&vsb_raw->apfs_fs_alloc_count, -1);
+		le64_add_cpu(&vsb_raw->apfs_total_blocks_freed, 1);
 		return 0;
 	case APFS_QUERY_OMAP:
 	case APFS_QUERY_EXTENTREF:
 	case APFS_QUERY_SNAP_META:
 		err = apfs_free_queue_insert(sb, bno, 1);
-		if (err)
+		if (err) {
+			apfs_err(sb, "free queue insertion failed for 0x%llx", bno);
 			return err;
+		}
 		/* We don't write to the container's omap */
 		vsb_raw = APFS_SB(sb)->s_vsb_raw;
 		apfs_assert_in_transaction(sb, &vsb_raw->apfs_o);
 		le64_add_cpu(&vsb_raw->apfs_fs_alloc_count, -1);
+		le64_add_cpu(&vsb_raw->apfs_total_blocks_freed, 1);
 		return 0;
 	case APFS_QUERY_FREE_QUEUE:
 		err = apfs_cpoint_data_free(sb, bno);
-		if (err)
+		if (err) {
+			apfs_err(sb, "failed to free checkpoint block 0x%llx", bno);
 			return err;
+		}
 		err = apfs_remove_cpoint_map(sb, bno);
-		if (err)
+		if (err) {
+			apfs_err(sb, "checkpoint map removal failed (0x%llx)", bno);
 			return err;
+		}
 		return 0;
 	default:
-		apfs_debug(sb, "new query type must implement node deletion");
+		apfs_alert(sb, "new query type must implement node deletion (%d)", query->flags & APFS_QUERY_TREE_MASK);
 		return -EOPNOTSUPP;
 	}
 }
@@ -517,8 +564,10 @@ int apfs_node_locate_key(struct apfs_node *node, int index, int *off)
 	struct apfs_btree_node_phys *raw;
 	int len;
 
-	if (index >= node->records)
+	if (index >= node->records) {
+		apfs_err(sb, "index out of bounds (%d of %d)", index, node->records);
 		return 0;
+	}
 
 	raw = (struct apfs_btree_node_phys *)node->object.data;
 	if (apfs_node_has_fixed_kv_size(node)) {
@@ -545,7 +594,7 @@ int apfs_node_locate_key(struct apfs_node *node, int index, int *off)
 	}
 
 	if (*off + len > sb->s_blocksize) {
-		/* Avoid out-of-bounds read if corrupted */
+		apfs_err(sb, "key out of bounds (%d-%d)", *off, len);
 		return 0;
 	}
 	return len;
@@ -568,8 +617,10 @@ static int apfs_node_locate_data(struct apfs_node *node, int index, int *off)
 	struct apfs_btree_node_phys *raw;
 	int len;
 
-	if (index >= node->records)
+	if (index >= node->records) {
+		apfs_err(sb, "index out of bounds (%d of %d)", index, node->records);
 		return 0;
+	}
 
 	raw = (struct apfs_btree_node_phys *)node->object.data;
 	if (apfs_node_has_fixed_kv_size(node)) {
@@ -613,7 +664,7 @@ static int apfs_node_locate_data(struct apfs_node *node, int index, int *off)
 	}
 
 	if (*off < 0 || *off + len > sb->s_blocksize) {
-		/* Avoid out-of-bounds read if corrupted */
+		apfs_err(sb, "value out of bounds (%d-%d)", *off, len);
 		return 0;
 	}
 	return len;
@@ -705,14 +756,12 @@ static int apfs_key_from_query(struct apfs_query *query, struct apfs_key *key)
 		err = apfs_read_omap_snap_key(raw_key, query->key_len, key);
 		break;
 	default:
-		/* Not implemented yet */
-		err = -EINVAL;
+		apfs_alert(sb, "new query type must implement key reads (%d)", query->flags & APFS_QUERY_TREE_MASK);
+		err = -EOPNOTSUPP;
 		break;
 	}
-	if (err) {
-		apfs_alert(sb, "bad node key in block 0x%llx",
-			   query->node->object.block_nr);
-	}
+	if (err)
+		apfs_err(sb, "bad node key in block 0x%llx", query->node->object.block_nr);
 
 	/* A multiple query must ignore some of these fields */
 	if (query->flags & APFS_QUERY_ANY_NAME)
@@ -749,21 +798,27 @@ static int apfs_node_next(struct super_block *sb, struct apfs_query *query)
 	query->key_len = apfs_node_locate_key(node, query->index,
 					      &query->key_off);
 	err = apfs_key_from_query(query, &curr_key);
-	if (err)
+	if (err) {
+		apfs_err(sb, "bad key for index %d", query->index);
 		return err;
+	}
 
 	cmp = apfs_keycmp(&curr_key, query->key);
 
-	if (cmp > 0) /* Records are out of order */
+	if (cmp > 0) {
+		apfs_err(sb, "records are out of order");
 		return -EFSCORRUPTED;
+	}
 
 	if (cmp != 0 && apfs_node_is_leaf(node) &&
 	    query->flags & APFS_QUERY_EXACT)
 		return -ENODATA;
 
 	query->len = apfs_node_locate_data(node, query->index, &query->off);
-	if (query->len == 0)
+	if (query->len == 0) {
+		apfs_err(sb, "bad value for index %d", query->index);
 		return -EFSCORRUPTED;
+	}
 
 	if (cmp != 0) {
 		/*
@@ -829,8 +884,10 @@ int apfs_node_query(struct super_block *sb, struct apfs_query *query)
 		query->key_len = apfs_node_locate_key(node, query->index,
 						      &query->key_off);
 		err = apfs_key_from_query(query, &curr_key);
-		if (err)
+		if (err) {
+			apfs_err(sb, "bad key for index %d", query->index);
 			return err;
+		}
 
 		cmp = apfs_keycmp(&curr_key, query->key);
 		if (cmp == 0 && !(query->flags & APFS_QUERY_MULTIPLE))
@@ -885,8 +942,10 @@ int apfs_omap_map_from_query(struct apfs_query *query, struct apfs_omap_map *map
 	struct apfs_omap_val *val = NULL;
 	char *raw = query->node->object.data;
 
-	if (query->len != sizeof(*val) || query->key_len != sizeof(*key))
+	if (query->len != sizeof(*val) || query->key_len != sizeof(*key)) {
+		apfs_err(sb, "bad length of key (%d) or value (%d)", query->key_len, query->len);
 		return -EFSCORRUPTED;
+	}
 	key = (struct apfs_omap_key *)(raw + query->key_off);
 	val = (struct apfs_omap_val *)(raw + query->off);
 
@@ -924,13 +983,17 @@ static int apfs_btree_inc_height(struct apfs_query *query)
 	root_raw = (void *)root->object.data;
 	apfs_assert_in_transaction(sb, &root_raw->btn_o);
 
-	if (query->parent || query->depth)
+	if (query->parent || query->depth) {
+		apfs_err(sb, "invalid root query");
 		return -EFSCORRUPTED;
+	}
 
 	/* Create a new child node */
 	new_node = apfs_create_node(sb, storage);
-	if (IS_ERR(new_node))
+	if (IS_ERR(new_node)) {
+		apfs_err(sb, "node creation failed");
 		return PTR_ERR(new_node);
+	}
 	new_node->flags = root->flags & ~APFS_BTNODE_ROOT;
 	new_node->tree_type = root->tree_type;
 
@@ -966,8 +1029,10 @@ static int apfs_btree_inc_height(struct apfs_query *query)
 	/* Now assemble the new root with only the first key */
 	root_query->key_len = apfs_node_locate_key(root, 0 /* index */,
 						   &root_query->key_off);
-	if (!root_query->key_len)
+	if (!root_query->key_len) {
+		apfs_err(sb, "bad key for index %d", 0);
 		return -EFSCORRUPTED;
+	}
 	root->key = sizeof(*root_raw) +
 		    apfs_node_min_table_size(sb, root->tree_type, root->flags & ~APFS_BTNODE_LEAF);
 	memmove((void *)root_raw + root->key,
@@ -1001,13 +1066,14 @@ static int apfs_btree_inc_height(struct apfs_query *query)
 }
 
 /**
- * apfs_copy_record_range - Copy a range of records to an empty nonroot node
+ * apfs_copy_record_range - Copy a range of records to an empty node
  * @dest_node:	destination node
  * @src_node:	source node
  * @start:	index of first record in range
  * @end:	index of first record after the range
  *
- * Returns 0 on success or a negative error code in case of failure.
+ * Doesn't modify the info footer of root nodes. Returns 0 on success or a
+ * negative error code in case of failure.
  */
 static int apfs_copy_record_range(struct apfs_node *dest_node,
 				  struct apfs_node *src_node,
@@ -1025,7 +1091,6 @@ static int apfs_copy_record_range(struct apfs_node *dest_node,
 	src_raw = (void *)src_node->object.data;
 
 	ASSERT(!dest_node->records);
-	ASSERT(!apfs_node_is_root(dest_node));
 	apfs_assert_in_transaction(sb, &dest_raw->btn_o);
 
 	/* Resize the table of contents so that all the records fit */
@@ -1038,7 +1103,9 @@ static int apfs_copy_record_range(struct apfs_node *dest_node,
 		toc_size = toc_entry_size * round_up(end - start, APFS_BTREE_TOC_ENTRY_INCREMENT);
 	dest_node->key = sizeof(*dest_raw) + toc_size;
 	dest_node->free = dest_node->key;
-	dest_node->data = sb->s_blocksize; /* Nonroot */
+	dest_node->data = sb->s_blocksize;
+	if (apfs_node_is_root(dest_node))
+		dest_node->data -= sizeof(struct apfs_btree_info);
 
 	/* We'll use a temporary query structure to move the records around */
 	query = apfs_alloc_query(dest_node, NULL /* parent */);
@@ -1052,8 +1119,10 @@ static int apfs_copy_record_range(struct apfs_node *dest_node,
 		int len, off;
 
 		len = apfs_node_locate_key(src_node, i, &off);
-		if (dest_node->free + len > sb->s_blocksize)
+		if (dest_node->free + len > sb->s_blocksize) {
+			apfs_err(sb, "key of length %d doesn't fit", len);
 			goto fail;
+		}
 		memcpy((char *)dest_raw + dest_node->free,
 		       (char *)src_raw + off, len);
 		query->key_off = dest_node->free;
@@ -1062,8 +1131,10 @@ static int apfs_copy_record_range(struct apfs_node *dest_node,
 
 		len = apfs_node_locate_data(src_node, i, &off);
 		dest_node->data -= len;
-		if (dest_node->data < 0)
+		if (dest_node->data < 0) {
+			apfs_err(sb, "value of length %d doesn't fit", len);
 			goto fail;
+		}
 		memcpy((char *)dest_raw + dest_node->data,
 		       (char *)src_raw + off, len);
 		query->off = dest_node->data;
@@ -1095,8 +1166,11 @@ static int apfs_attach_child(struct apfs_query *query, struct apfs_node *child)
 	__le64 raw_oid = cpu_to_le64(object->oid);
 
 	key_len = apfs_node_locate_key(child, 0, &key_off);
-	if (!key_len) /* This should never happen: @child was made by us */
+	if (!key_len) {
+		/* This should never happen: @child was made by us */
+		apfs_alert(object->sb, "bad key for index %d", 0);
 		return -EFSCORRUPTED;
+	}
 
 	return apfs_btree_insert(query, (void *)raw + key_off, key_len,
 				 &raw_oid, sizeof(raw_oid));
@@ -1154,9 +1228,12 @@ int apfs_node_split(struct apfs_query *query)
 
 	if (apfs_node_is_root(query->node)) {
 		err = apfs_btree_inc_height(query);
-		if (err)
+		if (err) {
+			apfs_err(sb, "failed to increase tree height");
 			return err;
+		}
 	} else if (!query->parent) {
+		apfs_err(sb, "nonroot node with no parent");
 		return -EFSCORRUPTED;
 	}
 	old_node = query->node;
@@ -1183,8 +1260,10 @@ int apfs_node_split(struct apfs_query *query)
 	old_node->key_free_list_len = 0;
 	old_node->val_free_list_len = 0;
 	err = apfs_copy_record_range(old_node, tmp_node, 0, old_rec_count);
-	if (err)
+	if (err) {
+		apfs_err(sb, "record copy failed");
 		goto out;
+	}
 	apfs_update_node(old_node);
 
 	/* The second half is copied to a new node */
@@ -1196,6 +1275,7 @@ int apfs_node_split(struct apfs_query *query)
 
 		new_node = apfs_create_node(sb, storage);
 		if (IS_ERR(new_node)) {
+			apfs_err(sb, "node creation failed");
 			err = PTR_ERR(new_node);
 			goto out;
 		}
@@ -1206,6 +1286,7 @@ int apfs_node_split(struct apfs_query *query)
 		new_node->val_free_list_len = 0;
 		err = apfs_copy_record_range(new_node, tmp_node, old_rec_count, record_count);
 		if (err) {
+			apfs_err(sb, "record copy failed");
 			apfs_node_free(new_node);
 			goto out;
 		}
@@ -1216,6 +1297,7 @@ int apfs_node_split(struct apfs_query *query)
 
 		err = apfs_attach_child(query->parent, new_node);
 		if (err) {
+			apfs_err(sb, "child attachment failed");
 			apfs_node_free(new_node);
 			goto out;
 		}
@@ -1407,14 +1489,18 @@ static int apfs_node_free_list_alloc(struct apfs_node *node, u16 len, bool value
 
 		if (curr_off == APFS_BTOFF_INVALID)
 			return -ENOSPC;
-		if (abs_off + sizeof(*curr) > sb->s_blocksize)
+		if (abs_off + sizeof(*curr) > sb->s_blocksize) {
+			apfs_err(sb, "nloc out of bounds (%d-%d)", abs_off, (int)sizeof(*curr));
 			return -EFSCORRUPTED;
+		}
 		curr = (void *)node_raw + abs_off;
 
 		curr_len = le16_to_cpu(curr->len);
 		if (curr_len >= len) {
-			if (abs_off + curr_len > sb->s_blocksize)
+			if (abs_off + curr_len > sb->s_blocksize) {
+				apfs_err(sb, "entry out of bounds (%d-%d)", abs_off, curr_len);
 				return -EFSCORRUPTED;
+			}
 			*list_len -= curr_len;
 			apfs_node_free_list_unlink(prev, curr);
 			apfs_node_free_list_add(node, abs_off + len, curr_len - len);
@@ -1425,6 +1511,7 @@ static int apfs_node_free_list_alloc(struct apfs_node *node, u16 len, bool value
 	}
 
 	/* Don't loop forever if the free list is corrupted and doesn't end */
+	apfs_err(sb, "free list never ends");
 	return -EFSCORRUPTED;
 }
 
@@ -1570,6 +1657,49 @@ fail:
 	return err;
 }
 
+/**
+ * apfs_node_total_room - Total free space in a node
+ * @node: the node
+ */
+static int apfs_node_total_room(struct apfs_node *node)
+{
+	return node->data - node->free + node->key_free_list_len + node->val_free_list_len;
+}
+
+/**
+ * apfs_defragment_node - Make all free space in a node contiguous
+ * @node: node to defragment
+ *
+ * Returns 0 on success or a negative error code in case of failure.
+ */
+static int apfs_defragment_node(struct apfs_node *node)
+{
+	struct super_block *sb = node->object.sb;
+	struct apfs_btree_node_phys *node_raw = (void *)node->object.data;
+	struct apfs_node *tmp_node = NULL;
+	int record_count, err;
+
+	apfs_assert_in_transaction(sb, &node_raw->btn_o);
+
+	/* Put all records in a temporary in-memory node and deal them out */
+	err = apfs_node_temp_dup(node, &tmp_node);
+	if (err)
+		return err;
+	record_count = node->records;
+	node->records = 0;
+	node->key_free_list_len = 0;
+	node->val_free_list_len = 0;
+	err = apfs_copy_record_range(node, tmp_node, 0, record_count);
+	if (err) {
+		apfs_err(sb, "record copy failed");
+		goto fail;
+	}
+	apfs_update_node(node);
+fail:
+	apfs_node_free(tmp_node);
+	return err;
+}
+
 /**
  * apfs_node_insert - Insert a new record in a node
  * @query:	query run to search for the record
@@ -1586,12 +1716,19 @@ fail:
 int apfs_node_insert(struct apfs_query *query, void *key, int key_len, void *val, int val_len)
 {
 	struct apfs_node *node = query->node;
+	struct super_block *sb = node->object.sb;
 	struct apfs_btree_node_phys *node_raw = (void *)node->object.data;
 	int toc_entry_size;
 	int old_free, old_data, old_key_free_len, old_val_free_len;
-	int key_off, val_off, err = 0;
+	int key_off, val_off, err;
+	int needed_room;
+	bool defragged = false;
 
-	apfs_assert_in_transaction(node->object.sb, &node_raw->btn_o);
+	apfs_assert_in_transaction(sb, &node_raw->btn_o);
+
+retry:
+	needed_room = key_len + val_len;
+	err = 0;
 
 	if (apfs_node_has_fixed_kv_size(node))
 		toc_entry_size = sizeof(struct apfs_kvoff);
@@ -1609,8 +1746,11 @@ int apfs_node_insert(struct apfs_query *query, void *key, int key_len, void *val
 
 		new_key_base += inc;
 		new_free_base += inc;
-		if (new_free_base > node->data)
-			return -ENOSPC;
+		if (new_free_base > node->data) {
+			needed_room += inc;
+			err = -ENOSPC;
+			goto fail_defrag;
+		}
 		memmove((void *)node_raw + new_key_base,
 			(void *)node_raw + node->key, node->free - node->key);
 
@@ -1627,14 +1767,14 @@ int apfs_node_insert(struct apfs_query *query, void *key, int key_len, void *val
 	key_off = apfs_node_alloc_key(node, key_len);
 	if (key_off < 0) {
 		err = key_off;
-		goto fail;
+		goto fail_update;
 	}
 
 	if (val) {
 		val_off = apfs_node_alloc_val(node, val_len);
 		if (val_off < 0) {
 			err = val_off;
-			goto fail;
+			goto fail_update;
 		}
 	}
 
@@ -1653,7 +1793,7 @@ int apfs_node_insert(struct apfs_query *query, void *key, int key_len, void *val
 	/* Add the new entry to the table of contents */
 	apfs_create_toc_entry(query);
 
-fail:
+fail_update:
 	/*
 	 * We must update the on-disk node even on failure, because we did
 	 * expand the table of contents.
@@ -1665,6 +1805,26 @@ fail:
 		node->val_free_list_len = old_val_free_len;
 	}
 	apfs_update_node(node);
+fail_defrag:
+	if (err == -ENOSPC && !defragged && needed_room <= apfs_node_total_room(node)) {
+		/*
+		 * If we know we should have enough room, defragment the node
+		 * and retry the insertion to avoid a split. This is especially
+		 * important for the ip free queue, because it has a very low
+		 * hard limit in the number of nodes allowed.
+		 */
+		err = apfs_defragment_node(node);
+		if (err) {
+			apfs_err(sb, "failed to defragment node");
+			return err;
+		}
+		defragged = true;
+		goto retry;
+	}
+	if (err == -ENOSPC && defragged) {
+		apfs_err(sb, "node reports incorrect free space");
+		err = -EFSCORRUPTED;
+	}
 	return err;
 }
 
@@ -1695,15 +1855,17 @@ int apfs_create_single_rec_node(struct apfs_query *query, void *key, int key_len
 
 	/* This function should only be needed for huge catalog records */
 	if (prev_node->tree_type != APFS_OBJECT_TYPE_FSTREE) {
-		apfs_warn(sb, "huge node records in the wrong tree");
+		apfs_err(sb, "huge node records in the wrong tree");
 		return -EFSCORRUPTED;
 	}
 
 	apfs_btree_change_node_count(query->parent, 1 /* change */);
 
 	new_node = apfs_create_node(sb, apfs_query_storage(query));
-	if (IS_ERR(new_node))
+	if (IS_ERR(new_node)) {
+		apfs_err(sb, "node creation failed");
 		return PTR_ERR(new_node);
+	}
 	new_node->tree_type = prev_node->tree_type;
 	new_node->flags = prev_node->flags;
 	new_node->records = 0;
@@ -1725,7 +1887,12 @@ int apfs_create_single_rec_node(struct apfs_query *query, void *key, int key_len
 	query->node = new_node;
 	query->index = -1;
 	err = apfs_node_insert(query, key, key_len, val, val_len);
-	if (err)
+	if (err) {
+		apfs_err(sb, "node record insertion failed");
 		return err;
-	return apfs_attach_child(query->parent, new_node);
+	}
+	err = apfs_attach_child(query->parent, new_node);
+	if (err)
+		apfs_err(sb, "child attachment failed");
+	return err;
 }
